#!/usr/bin/env perl
# Description: Analyse network traffic (dump file).
#
# Copyright:
# Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

# mapwocanalyse.pl - Analayse network traffic from dump file.
# Usage:
#   mapwocanalyse -p <dump_file> -b [working_directory] -j [json_file]
# working_directory defaults to cwd.

# Result:
# Creates working_directory/Log and working_directory/Bin.

use strict;
use warnings;

use Cwd;
use File::Spec;
use File::Basename;
use Getopt::Std;

use MapWOCAnalyser::MapWOCAnalyser qw( @filename ); # import @filename from MapWOCAnalyser.pm

my %opts;

getopt('b:p:j:', \%opts); # get arguments

my $BASE_DIR = cwd; # Set BASE_DIR to current dir.

if ( defined $opts{ 'b' } ) { # Use another BASE_DIR?

    if( ! -d $opts{ 'b' } ) { 
        if( ! mkdir ("$opts{ 'b' }", 0777) ) {
            print "ERROR: Could not create base-dir '" . $opts{ 'b' } . "'Exit now!\n";
            exit 252;
        }
    }

    $BASE_DIR = $opts{ 'b' };
}

if ( ! chdir( $BASE_DIR ) ) {
    print "ERROR: Could not change into dir $BASE_DIR\n";
    exit 254;
}

if( ! defined $opts{ 'p' } ) { # pcapfile
    print "Usage: mapwocanalyse -p <dump_file> -b [working_directory] -j [json_file]\n";
    exit 255;
}

if( ! File::Spec->file_name_is_absolute( $opts{ 'p' } ) ) { # path to pcap file is absolut?
    print "ERROR: Path to dump file must be absolute.\n";
    exit 243;
}

my $TRUE  = 1;
my $FALSE = 0;

my $err;

my $LOG_DIR    = $BASE_DIR . '/Log';
my $BINARY_DIR = $BASE_DIR . '/Binary';
my $xmlfile    = $LOG_DIR  . '/Analyse.xml';

if( ! -d $LOG_DIR ) { # Create dir for logs if it does not exist.
    if( ! mkdir ("$LOG_DIR", 0777) ) {
        print "ERROR: Could nor create logdir $LOG_DIR. Exit now!\n";
        exit 255;
    }
}

if( ! -d $BINARY_DIR ) { # Create dir for binarys if it does not exist.
    if( ! mkdir ("$BINARY_DIR", 0777) ) {
        print "ERROR: Could nor create binarydir $BINARY_DIR. Exit now!\n";
        exit 255;
    }
}

my $pcapfile = $opts{ 'p' };

if ( ! -e $pcapfile ) {
    print "ERROR: PCAP File $pcapfile not found. Exit now!\n";
    exit 255;
}

my $json_file = undef;

if ( defined ( $opts{ 'j' } ) ) {

    $json_file = $opts{ 'j' };

    if ( ! -e $json_file ) {
        print "ERROR: JSON File $json_file not found. Exit now!\n";
        exit 255;
    }

}

# Create dump dir from pcap filename.
my @path = split( /\//, $pcapfile );
my @work = split( /\./, $path[( scalar @path) -1] );
$work[ ( scalar @work ) -2 ] =~ s/:/_/mg;

my $work_dir = $work[( scalar @work ) -2 ]; # workdir based on pcap file name

if( ! -d $work_dir ) {
    if( ! mkdir ("$work_dir", 0777) ) {
        print "ERROR: Could nor create workdir $work_dir. Exit now!\n";
        exit 255;
    }
}

#########################################################################

# create a new analyser object.

my $analyser = MapWOCAnalyser::MapWOCAnalyser->_new(
                                                    BINARY_DIR  => $BINARY_DIR,
                                                    pcapfile    => $pcapfile,
                                                    work_dir    => $work_dir,
                                                    LOG_DIR     => $LOG_DIR,
                                                    xmlfile     => $xmlfile,
                                                    json_file   => $json_file
                                              );

{

    $/ = undef;
    $\ = "";
    $| = 0;

}

$analyser->{ 'XML' }->_start_xml( basename ( $pcapfile ) );

foreach my $name ( @filename ) { # @filename is defined in callback function 'syn_packets()' in MapWOCAnalyser.pm

    print "INFO: Evaluating $BASE_DIR/$name\n";

    if ( ! $analyser->_read_traffic( $BASE_DIR . "/" . $name ) ) { # read traffic

        print "ERROR: read_traffic() for  $BASE_DIR . "/" . $name failed. Try to continue...\n";
        next;

    }
}

if( defined ( $json_file ) ) {
    $analyser->_read_json( $json_file );
}

my @antivir_tmp = ( $BINARY_DIR, $BASE_DIR . "/" . $work_dir ); # Check both. Binary and Dump dir...

foreach my $dir ( @antivir_tmp )  {

    my @filesFound = glob( $dir . "/*" );
    my $file_count = @filesFound;

    if ( $file_count > 0 ) {

        my $scanner = "/usr/bin/clamscan";

        if ( -e $scanner && -x $scanner ) {

            print "Scanning files in dir $dir for Viruses. Scanner is: $scanner\n";

            my $antivir     = qx( $scanner $dir );
            my @antivir_out = split( /\n/, $antivir );

            foreach my $av_line ( @antivir_out ) {

                $av_line =~ s/[\r\n]+//g;
                last if( $av_line =~ /^$/ );

                if( $av_line =~ m/FOUND$/ ) {

                    my @av_split = split( / /, $av_line );
                    $analyser->{ 'XML' }->_add_antivir_result_to_xml( @av_split ); # update antivir placeholders with antivir result

                }

            }

        } else {

            print "Scanner $scanner does not exist. Exit now!.\n";
            exit 255;

        }
    }
}

$analyser->{ 'XML' }->_close_xml;

__END__

=head1 NAME

mapwocanalyse.pl - Perl Script for mapwoc

=head1 DESCRIPTION

Script for running the mapwoc analyser. 

=head1 LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2,
or, at your option, any later version as published by the Free
Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

=head1 AUTHOR

Michael Meyer <michael.meyer@intevation.de>

=head1 COPYRIGHT

Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik

=cut
