# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Convenience- methods and constants for mapwoc applications.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
import logging
from optparse import OptionParser

BSI_COPYRIGHT = ("Copyright (C) 2011 Bundesamt für"
                 " Sicherheit in der Informationstechnik, BSI")
VERSION_STATEMENT = ("mapwoc comes with ABSOLUTELY NO WARRANTY;\n"
                     "This is free software, and you are welcome to\n"
                     "redistribute it under certain conditions (GPLv2+).")

def print_version_exit(tool, version):
    """
    Print tool name, version, copyright, free software
    statement and exit(0).
    """
    print "%s version %s, %s\n%s" % (tool, version, BSI_COPYRIGHT,
                                     VERSION_STATEMENT)
    exit(0)

def setup_logger(quiet, verbose, loggername, logfile):
    """Setup logging."""

    logger = logging.getLogger(loggername)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    if verbose:
        min_level = logging.DEBUG
    elif quiet:
        min_level = logging.WARNING
    else:
        min_level = logging.INFO

    lfh = logging.FileHandler(logfile)
    lsh = logging.StreamHandler()
    lfh.setFormatter(formatter)
    lsh.setFormatter(formatter)

    logger.addHandler(lfh)
    logger.addHandler(lsh)
    logger.setLevel(min_level)

"""
To ease cross-application consistent behaviour with respect to
configuration files and command line (configuration) options,
the mapwoc.lib.toolutil- module provides two classes to enable
the applications to be configured.

An option is specified like:
(short_option, long_option, description, default, type) .
"""


class MapwocOptionParser(OptionParser):
    COMBIOPT_TYPE_INT = 'int'
    COMBIOPT_TYPE_FLOAT = 'float'
    COMBIOPT_TYPE_STRING = 'string'
    COMBIOPT_TYPE_MAKETRUE = 'settrue'

    def __init__(self, version, description,
                 default_config_file_path,
                 combiopts=[], allow_gen=True):
        """
        Initialalizes internal OptionParser.
        @param version: Version String to display when queried (--version).
        @param default_config_file_path: Path to the (default) configuration
                                         file.
        @param combiopts: list of combiopts.
        @param allow_gen: Whether or not to allow to create a config file
                          with --generate-config.
        """
        OptionParser.__init__(self,
                              version=version,
                              description=description)
        self.combiopts = combiopts
        if not self.combiopts:
            self.combiopts = []
        self.default_config_file_path = default_config_file_path
        self.combiopts.append(('', 'config', 'Path to configuration file.',
                               default_config_file_path, self.COMBIOPT_TYPE_STRING),)
        if allow_gen:
            self.combiopts.append(('', 'generate-config', 'Generate a configuration file.',
                                   False , self.COMBIOPT_TYPE_MAKETRUE),)

        self.inject_combiopts(self.combiopts) 

    def inject_combiopts(self, combiopts):
        """
        Inject options into the OptionParser.
        @param combiopts: List of option-tuples.
        """
        for (oshort,olong,odesc,odef,otype) in combiopts:
            short_opt = oshort
            if short_opt and short_opt != '':
                short_opt = '-%s' % short_opt
            long_opt = '--%s' % olong
            if otype == self.COMBIOPT_TYPE_INT:
                self.add_option(short_opt, long_opt, dest=olong, help=odesc,
                                type=int)
            elif otype == self.COMBIOPT_TYPE_FLOAT:
                self.add_option(short_opt, long_opt, dest=olong, help=odesc,
                                type=float)
            elif otype == self.COMBIOPT_TYPE_MAKETRUE:
                self.add_option(short_opt, long_opt, dest=olong, help=odesc,
                                action="store_true")
            else:
                self.add_option(short_opt, long_opt, dest=olong, help=odesc) 

    def parse_args(self):
        (options, args) = OptionParser.parse_args(self)
        # Config items given at cmd override config options given in file,
        # which override defaults.
        conf_defaults = {}
        for (__oshort,olong,__odesc,odef,__otype) in self.combiopts:
            conf_defaults[olong] = odef

        config = ConfigParser.RawConfigParser(defaults=conf_defaults)
        try:
            if options.config == None:
                options.config = self.default_config_file_path
            config.read(options.config)
            if not config.has_section('config'):
                config.add_section('config')

            for (oshort,olong,odesc,odef,otype) in self.combiopts:
                cmd_opt_val = getattr(options, olong)
                if not cmd_opt_val:
                    if otype == self.COMBIOPT_TYPE_INT:
                        cfg_opt_val = config.getint('config', olong)
                    elif otype == self.COMBIOPT_TYPE_FLOAT:
                        cfg_opt_val = config.getfloat('config', olong)
                    else:
                        cfg_opt_val = config.get('config', olong)
                    setattr(options, olong, cfg_opt_val)
        except ConfigParser.NoSectionError:
            print "Given configuration file %s is invalid or missing option." % options.config
            exit(2)
        except TypeError:
            # TODO It can be checked whether the configuration file exists or not.
            if getattr(options, 'generate-config'):
                print "Generating a config in %s" % options.config
                if not config.has_section('config'):
                    config.add_section('config')
                for (oshort,olong,odesc,odef,otype) in self.combiopts:
                    config.set('config', olong, odef)
                optf = open(options.config, 'w')
                config.write(optf)
                print "Config file written."
            else:
                print "Error parsing the options, does the config file exist?"

            exit(3)

        return (options, args)

