# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Master-centric classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
# Andre Heinecke <aheinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
import datetime
import json
import logging
from multiprocessing import Process, Queue
from os import listdir, path, makedirs
from Queue import Empty
import subprocess
import socket
import time
from time import sleep

from sqlalchemy import *

from mapwocgui.model.meta import Session

from mapwoc.lib.trivialresponse import TrivialResponse
from mapwoc.lib.message import Message
from mapwoc.lib.mailhelper import MailDispatcher
from mapwoc.lib.users import UserData
from mapwoc.lib.hc import HC
from mapwoc.lib.node import Node
from mapwoc.lib.scanconfiguration import ScanConfiguration
from mapwoc.lib.scanconfiguration import ScanRun
from mapwoc.lib.scanconfiguration import TransportableSC
from mapwoc.lib.scanconfiguration import VisitorConfiguration
from mapwoc.lib.scanconfiguration import VisitorBundle

from nodeconnection import NodeConnection
from nodepollprocess import NodePollProcess
from redirectorpollprocess import ask_redirector, ask_redirector_decode
from redirectorpollprocess import RedirectorPollProcess
from resultpollprocess import ResultPollProcess

from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook
from mapwoc.lib.listdict import listdict

logger = logging.getLogger('master')

class CannotStartScanError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class MasterState(object):
    """Representation of the state of a master."""

    IDLE,SCANNING,STOPPING,ERROR = range(1,5)
    strstate = {IDLE : 'idle',
                SCANNING : 'scanning',
                STOPPING : 'stopping',
                ERROR : 'error'}

    def __init__(self):
        """Initialiases Master State Set."""

        self.state = self.IDLE
        self.scanname = None
        self.nodestates = {}
        self.nodestates_are_valid = True
        self.redirectorstate = None

    def set(self, newstate):
        """Sets state of master."""

        self.state = newstate

    def export(self):
        """Returns dictionary of stateset as expected by mapwocgui."""

        return {'master' : self.strstate[self.state],
                'scanname' : self.scanname,
                'nodes' : self.nodestates,
                'redirector' : self.redirectorstate}

    def of_nodes_invalidate(self):
        """Invalidates the stateset of the nodes until they are updated."""
        self.nodestates_are_valid = False

    def of_nodes_is_valid(self):
        """Returns true if the nodestates are valid."""
        return self.nodestates_are_valid

    def of_nodes_set(self, newstates):
        """Sets states of nodes."""
        self.nodestates = newstates
        self.nodestates_are_valid = True

    def of_redirector_set(self, newstate):
        """Sets state of redirector."""
        self.redirectorstate = newstate

class Master:
    """A Master."""

    VERSION = "0.6.7"

    def __init__(self, configfile, node_config_dir, credentials_path,
                 work_dir, db, redirector_port, dry=False,
                 severity_tolerance=0):
        """
        Init fields to empty lists and reads configuration file.
        @param configfile: Path to the configuration file.
        @param node_config_dir: Path to the directory containing node information files.
        @param credentials_path: Path to the credentials to use when polling
                                 results from nodes.
        @param work_dir: Path to working directory.
        @param db: Path to database.
        @param redirector_port: Port number of redirector (for commands).
        @param dry: Set to True to keep subprocesses unstarted (to debug).
        @param severity_tolerance: Mimumum value of results to store.
        """

        self.__nodes = {}  # Mapping node keys (names) to NodeConnections
        self.messages = [] # List of Messages
        self.__hcs = [] # List of HC's
        self.__scan = None # ScanRun Instance
        
        self.__read_conf(configfile)
        self.node_config_dir = node_config_dir
        self.credentials_path = credentials_path
        self.__work_dir = work_dir
        self.__db_file = db
        self.redirector_port = redirector_port
        self.__load_node_config()
        self.__sync_hcs()

        # Setup and start child processes that poll the other services.
        self.node_queue = Queue()
        self.node_poll_process = NodePollProcess(self.node_queue,
                                                 self.__nodes,
                                                 5)
        if not dry:
            self.node_poll_process.start()

        self.redirector_queue = Queue()
        self.redirector_poll_process = RedirectorPollProcess(
                                            self.redirector_queue, 5,
                                            port = self.redirector_port)
        if not dry:
            self.redirector_poll_process.start()

        if not path.isdir(self.__work_dir):
            makedirs(self.__work_dir)

        engine = create_engine('sqlite:///%s' % self.__db_file)
        __c = engine.connect()
        Session.configure(bind=engine)
                                
        self.resultsyncpoller = ResultPollProcess(self.__work_dir,
                                                  self.__nodes, Queue(),
                                                  self.credentials_path,
                                                  self.__db_file,
                                                  severity_tolerance)
        if not dry:
            self.resultsyncpoller.start()

        self.__state = MasterState()

    def __del__(self):
        """Deconstructor. Take care of child processes, open handles etc."""

        #TODO terminate and join the poll processes.
        pass

    def __update_redirector_state(self):
        """Sets state_set['redirector'] and places a Message."""
        try:
            # Get last answer to 'state' question of redirector.
            redirector_answer = self.redirector_queue.get_nowait()
            while not self.redirector_queue.empty():
                redirector_answer = self.redirector_queue.get_nowait()

            self.__state.of_redirector_set(redirector_answer)
        except Empty:
            logger.debug("No update from redirector state poll.")

    def __update_node_state(self):
        """Sets state_set['nodes']."""
        try:
           node_answer = self.node_queue.get_nowait()
           while not self.node_queue.empty():
               node_answer = self.node_queue.get_nowait()
           self.__state.of_nodes_set(node_answer)
        except Empty:
            logger.debug("No update from nodes state poll yet.")

    def check_state_polls(self):
        """
        Checks the Queues which are communication channels to the processes
        that ask related services about their state.  Update state.
        """

        self.__update_redirector_state()
        self.__update_node_state()

        self.__update_state()

    def __nodes_idle(self):
        """True if all nodes state is 'idle'."""

        if not self.__state.of_nodes_is_valid():
            return False

        for _node,state in self.__state.nodestates.items():
            if state.get('') != 'idle':
                return False
        return True

    def __update_status_scan(self):
        """Set self.__state.scanname."""

        if self.__scan:
            self.__state.scanname = self.__scan.scanconfiguration.name
        else:
            self.__state.scanname = None

    def __update_state(self):
        """
        Act upon state of all components.
        Note that the state information might be outdated.
        """

        # The DO-Matrix:
        # Master | Redirector | Node1 | Node2 | DO
        #------------------------------------------
        #  (1 of these error)                 | handle error
        # idle   | idle       | idle  | idle  | --
        # idle   | *          | run   | *     | handle_wild_node
        # idle   | (1 of these not idle)      | become non-idle
        # running| idle       | *     | *     | ?
        # running| done       | *     | *     | stop_nodes
        # running| 

        # TODO Handle 'error' states.
        if not self.__state.of_nodes_is_valid():
            logger.info("Invalidated node states (wait for next poll).")
            return

        # TODO If scan is a TimedScan, check its time, stop if necessary.

        if self.__scan is None:
            # Not Scanning.
            # TODO The other components should idle, too!
            if self.__state.of_nodes_is_valid() and not self.__nodes_idle():
                logger.warning("Node does not idle but master has no scan!")

            if not self.__state.state == MasterState.IDLE:
                # Become idle.
                self.__state.set(MasterState.IDLE)
                logger.info("Setting master state to idle.")
        else:
            # Scanning.

            # Timed scan (ignorant wrt redirector).
            if self.__scan.scanconfiguration.is_timed_scan():
                started = self.__scan.starttime
                duration = datetime.datetime.now() - started
                max_duration = datetime.timedelta(seconds=abs(
                                self.__scan.scanconfiguration.visitor_configuration.active_window))

                if duration > max_duration:
                    logger.debug("Stopping timed scan.")
                    # TODO _gracefully_ stop the timed scan, ignore the
                    #      redirector state (timed scans look like aborted).
                    for _name,node in self.__nodes.items():
                        node.ask('stopsmooth')
                    self.__state.set(MasterState.STOPPING)

            if self.__state.redirectorstate.get('state') == 'done' or self.__scan.scanconfiguration.is_timed_scan():
                if self.__nodes_idle():
                    # TODO Tell redirector to idle. Listens to what it says.
                    try:
                        ask_redirector('go_idle', 'localhost', self.redirector_port)
                    except:
                        logger.warning("Couldnt ask redirector to go idle.")
                    # Tell world that scan is done.
                    self.messages.append(Message(Message.LEVEL_FUNCTIONAL_LOG,
                                         "Scan %s done" % self.__scan.scanconfiguration.name,
                                         heading="Scan"))
                    self.__send_scan_finished(
                            self.__scan.scanconfiguration.name)
                    self.__scan.stoptime = datetime.datetime.now()
                    self.__scan.status = "done"
                    Session.add(self.__scan)
                    Session.commit()
                    self.__scan = None
                    self.__state.set(MasterState.IDLE)
                    logger.debug("Go idle (r. is idle, nodes are idle)")
                else:
                    if not self.__scan.scanconfiguration.is_timed_scan():
                        for _name,node in self.__nodes.items():
                            node.ask('stopsmooth')
                        self.__state.set(MasterState.STOPPING)
            elif self.__state.state == MasterState.STOPPING and self.__nodes_idle():
                # Tell world that scan is aborted.
                self.messages.append(Message(Message.LEVEL_FUNCTIONAL_LOG,
                                     "Scan %s aborted" % self.__scan.scanconfiguration.name,
                                     heading="Scan"))
                self.__scan.stoptime = datetime.datetime.now()
                self.__scan.status = "aborted"
                Session.add(self.__scan)
                Session.commit()
                self.__scan = None
                self.__state.set(MasterState.IDLE)
            elif self.__state.state != MasterState.STOPPING and not self.__nodes_idle():
                # Nodes are scanning and we are not stopping
                self.__state.set(MasterState.SCANNING)
            else:
                logger.error("Do not know how to act on state.")

        self.__update_status_scan()

        # TODO Spell out rules. Keep in mind that state information might be relatively old.
        # If self running/stopping and nodes idle and redirector idle or done, say scan done, become idle
        # If self running and redirector done, become stopping, make nodes stopping
        # If self running/stopping, nodes stopped, redirector done, idle the redirector
        # If self stopping, nodes running, redirector not done: abort scan, become idle

    def get_node_by_name(self, name):
        """ Get a Node object by it's name. """

        return self.__nodes[name]

    def get_nodes(self):
        """Accessor for nodes."""

        return self.__nodes

    def check_scan(self, scanconf):
        """
        Checks whether a scan could be started by the nodes.
        @param scanconf: The Scan Configuration to let the nodes check.
        @return: Triple, first value being True if all nodes could start that
                 scan.  Otherwise (False,nameofnode,messge).
        """

        nodetasks = listdict()

        # Sort single HCs into correct nodetasks pocket.
        for visitor in scanconf.visitor_configuration.visitors:
            node = self.__nodes.get(visitor.node)
            if node is None:
                logger.info("Node %s for which jobs exist not found." % \
                        visitor.node)
            else:
                nodetasks.lappend(visitor.node, visitor)

        for node, visitorlist in nodetasks.items():
            node_vconf = VisitorConfiguration()
            node_vconf.visitors = visitorlist
            node_vconf.active_window = scanconf.visitor_configuration.active_window
            n = self.__nodes.get(node)
            ok,tr = n.ask_ok_tr("checkjob\r\n%s" 
                                % MapWOCJSONEncoder().encode(node_vconf))
            if not ok:
                logger.warning("Node will not be able to do the job: %s." % tr.message)
                return (False,node,tr.message)

        return (True,None,None)

    def __node_start_scan(self, nodename, visitor_configuration,
                          visitorlist, scanconf):
        """
        Start a Scan on a node, return its response.
        @param nodename: Name of node to start scan on.
        @param visitor_configuration: Configuration to use.
        @param visitorlist: Visitorbundles to use.
        @param scanconf: The scanconfiguration to use.
        @return: Answer of node.
        """
        # TODO Further refactoring might make some of the parameters obsolete
        #      (e.g. visitor_conf and visitorlist can be taken from scanconf)
        node = self.__nodes.get(nodename)
        vconf = VisitorConfiguration()
        vconf.eventmap = visitor_configuration.eventmap
        vconf.active_window = visitor_configuration.active_window
        vconf.visitors = visitorlist
        job = TransportableSC(scanconf)
        job.visitor_configuration = vconf
        question = "startrun\r\n%d\r\n%s" % ( self.__scan.id,
                                  MapWOCJSONEncoder().encode(job))
        return node.ask_decode(question)

    def start_scan(self, scanconf_id):
        """
        Start a ScanRun with a given ScanConfiguration.

        Raises CannotStartScanError s.

        @param scanconf: The ScanConfiguration to distribute to nodes and
                         Redirector.
        @return: False in case of errors, True otherwise.
        """

        if self.__scan:
            logger.warning("Master can only have one scan running at a time.")
            raise CannotStartScanError("Already running a scan.")

        # TODO Respect own state.

        scanconf = Session.query(ScanConfiguration).get(scanconf_id)
        if not scanconf:
            logger.warning("Master did not find scanconf %d in database." % scanconf_id)
            raise CannotStartScanError("Scan %d not found in database." % scanconf_id)

        if scanconf is None:
            logger.warning("Given no ScanConfiguration.")
            return False
        vc = scanconf.visitor_configuration
        if vc is None:
            logger.warning("Given Scan Configuration lacks visitor configuration.")
            return False
        if not vc.visitors:
            logger.warning("Given Vistor Configuration does not have Visitor entries.")
            return False

        # Check whether all HCs are known and whether the types are not mixed.
        nr_native = 0
        nr_virtual = 0
        for visitor in vc.visitors:
            hc = self.get_hc(visitor.hcsid)
            if hc is None:
                raise CannotStartScanError("Scan contains an unknown HC.")
            if hc.hctype == 'native':
                if nr_virtual > 0:
                    raise CannotStartScanError("Scan cannot have mixed types of HCs (native/virtual).")
                nr_native = nr_native + 1
            else:
                if nr_native > 0:
                    raise CannotStartScanError("Scan cannot have mixed types of HCs (native/virtual).")
                nr_virtual = nr_virtual + 1

        (can,node,msg) = self.check_scan(scanconf)
        if not can:
            raise CannotStartScanError(msg)

        nodetasks = listdict()
        self.__scan = ScanRun()
        self.__scan.scanconfiguration_id = scanconf.id
        self.__scan.status = "start prepared"

        Session.add(self.__scan)
        Session.commit()

        if scanconf.is_timed_scan():
            logger.debug("Start a timed scan.")
        else:
            logger.debug("Start a regular scan.")

        collections=""
        # Create a colon seperated list of urlcollection ids.
        for urllist in scanconf.urlcollection:
            collections += "%d," % urllist.id
        # Give order to redirector.
        r_response = ''
        try:
            r_response = ask_redirector_decode('loaddb %s,%s' % (collections,
                path.abspath(self.__db_file)), 'localhost', self.redirector_port)
        except:
            logger.warning("Could not ask or decode redirectors answer"
                           " to loaddb question")
            r_response = ''

        if r_response == None:
            logger.warning(
                    "panic: Redirector did not answer to loaddb question")
        if r_response.error != 0 and not self.__scan.scanconfiguration.is_timed_scan():
            logger.warning("Redictor answered %s " % r_response.message)
            return False

        # Communicate amout of URLs to scan at once.
        try:
            r_response = ask_redirector('fastmode %d' % vc.num_urls,
                    'localhost', self.redirector_port)
        except:
            logger.warning("Could not ask redirector to go to fast mode.")
        logger.debug("Asking redirector to move to fastmode %d " % vc.num_urls)
        if r_response != 'ok':
            logger.warning(
                    "panic: Redirector did not acknowledged fastmode question")

        # Sort single HCs into correct nodetasks pocket.
        for visitor in vc.visitors:
            node = self.__nodes.get(visitor.node)
            if node is None:
                logger.info("Node %s for which jobs exist not found." % \
                        visitor.node)
            else:
                nodetasks.lappend(visitor.node, visitor)

        # Create visitorconfiguration, add to scanconf, send to node.
        for node, visitorlist in nodetasks.items():
            response = self.__node_start_scan(node, vc,
                                              visitorlist, scanconf)
            if response.error != 0:
                # TODO which state to go into?
                logger.warning("Node answered: %s" % response.message)
                return False
            else:
                self.__state.of_nodes_invalidate()
                logger.debug("Node accepted to start scan.")

        logger.info("Starting scan.")
        self.messages += [Message(Message.LEVEL_FUNCTIONAL_LOG,
                                  "Starting Scan %s" % scanconf.name,
                                  heading="Scan")]
        # Set State.
        self.__state.set(MasterState.SCANNING)
        self.__state.scanname = scanconf.name
        return True

    def stop_scan(self, graceful=False):
        """Stops a running scan."""

        if not self.__scan:
            logger.warning("No Scan found to stop.")
            return False
       
        for v in self.__scan.scanconfiguration.visitor_configuration.visitors:
            if graceful:
                response = self.__nodes[v.node].ask('stopsmooth')
            else:
                response = self.__nodes[v.node].ask('stoprun')

        try:
            ask_redirector('go_idle', 'localhost', self.redirector_port)
        except:
            logger.warning("Couldnt ask redirector to go idle.")

        # Collect the answers from the stop questions, remove the scan config if
        # successfull. (TODO)

        self.__state.set(MasterState.STOPPING)
        return True

    def state(self):
        """Returns string representation of state."""

        return self.__state.export()

    def __sync_hcs(self):
        """
        Syncs HCs with configured nodes.
        Calling might block for a longer time.
        """
        for key, node in self.__nodes.items():
            logger.debug("Getting hc's from: %s" % key)
            hcs = node.ask_decode('hcs')
            if not hcs:
                logger.warning("Node %s reports no HC's." % key)
                continue
            for hc in hcs:
                # Check if hcid is already known
                for known in self.__hcs:
                    if known.hcsid == hc.hcsid:
                        logger.debug("Adding known hc: %s " % hc.hcsid)
                        known.on_nodes.append(key)
                        break
                else:
                    logger.debug("Adding unknown hc: %s " % hc.hcsid)
                    hc.on_nodes = [key]
                    self.__hcs.append( hc )


    def __read_conf(self, config_file):
        """Reads in configuration from file."""

        config = ConfigParser.RawConfigParser ()
        config.read(config_file)
        try:
            self.__maildispatcher = MailDispatcher(
                    config.get('notifications', 'smtp_server'),
                    config.get('notifications', 'smtp_port'),
                    config.get('notifications', 'admin_mail'),
                    config.get('notifications', 'sender_mail'))
        except:
            logger.info("No valid mail configuration found. "
                        "Disabling Notifications.")
            self.__maildispatcher = MailDispatcher(
                    None, None, None, None)

    def __load_node_config(self):
        """Load the node config files from node config directory."""
       
        # TODO Think about refactoring this and move the config reading
        #      part somewhere else, just keep a register_node function.
        logger.info("Loading nodes from %s." % self.node_config_dir)

        confdir = self.node_config_dir
        # Expand and normalize all paths.
        abs_file_paths = [ path.normpath(path.join(confdir, file_rel))
                           for file_rel in listdir(confdir) ]
        # Just keep the ones ending in .conf .
        abs_conf_files = [ file_abs for file_abs in abs_file_paths
                           if file_abs.endswith(".config") ]

        # Parse the remaining ones and add to list of nodes.
        for node_conf_file in abs_conf_files:
            node = NodeConnection(node_conf_file)
            self.__nodes[node.key] = node
            node.data_path = node.ask('resultdir')
            if not node.data_path.startswith('/'):
                logger.error("Could not get resultsdir from node %s " % node.key)
                node.data_path = None

        logger.info("Loaded %d nodes." % len(self.__nodes))


    def get_hcsid_on_nodes(self):
        """
        @return: A mapping from hcsid to list of nodes.
        """

        answer = {}
        for hc in self.__hcs:
            for node in hc.on_nodes:
                if answer.get(hc.hcsid):
                    answer[hc.hcsid].append(node)
                else:
                    answer[hc.hcsid] = [node]
        return answer

    def get_hc(self, hcsid):
        """
        @return: HCControl belonging to hcsid.
        """
        for hc in self.__hcs:
            if hc.hcsid == hcsid:
                return hc
        return None


    def get_last_msgs(self, num=1):
        """
        Get the num last messages as a list.

        @param num: Number of messages to fetch.
        @return: The last num messages (first item in list is last message).
        """

        msgs = self.messages[-num:]
        msgs.reverse()
        return msgs

    def __send_scan_finished(self, scan_name):
        """
        Sends a email to all users who wanted to be notified on success.
        """
        for user_data in Session.query(UserData).all():
            if ( user_data.notify_on_success == True and
                 user_data.e_mail != None ):
                self.__maildispatcher.send_scan_finished(
                        scanname=str(scan_name),
                        reciever=str(user_data.e_mail))
