# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process results.
#
# Authors:
# Andre Heinecke <aheinecke@intevation.de>
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
from datetime import datetime
import logging
import os
import shutil
import tarfile
import traceback

import mapwoc.lib.util as util

from mapwoc.lib.result import Result, Visit, ResultType, ResultFile
from mapwoc.lib.result import static_result_type_id
from mapwoc.lib.result import static_result_type_severity
from mapwoc.lib.scanconfiguration import ScanConfiguration, ScanRun
from mapwocgui.model.meta import Session
from mapwoc.lib.url import URL

logger = logging.getLogger('master')

DEBUGANALYSER = True

class ResultAnalyser(object):
    """ Provides a base class for Result Analysers """

    def __init__(self, result_data, node, tolerance):
        """
        ResultAnalyser class to enter raw result_data into the Database as
        results.

        @param result_data: Qualifier (path to) of the raw data which should be parsed.
        @param node: Name of node on which the scan was running.
        @param tolerance: Required highest severity for a result to store.
        """
        if DEBUGANALYSER:
            logger.debug("ResultAnalyser __init__ with data %s for node %s" %\
                    (result_data, node))
        self.result_data = result_data
        self.visit = Visit()
        self.visit.filepath = result_data
        self.visit.node = node
        Session.autoflush = False
        self.tolerance = tolerance

    def scan_metainformation(self, fptr):
        """
        Scan a Meta Information file.

        @param fptr: Filelike object with meta information about Visit.
        """
        parser = ConfigParser.RawConfigParser()
        parser.readfp(fptr)

        self.visit.scanrun_id = parser.getint("scan", "scanrunid")
        self.visit.date = datetime.fromtimestamp(parser.getfloat("scan", "time"))

        scanrun = Session.query(ScanRun).filter_by(id=self.visit.scanrun_id).first()
        if DEBUGANALYSER:
            logger.debug("Metainformation scanrun.id %d" % self.visit.scanrun_id)
        self.visit.scanrun = scanrun

    def construct_result(self, url_string, type_name, method, date=None):
        """
        Construct a Result object.

        @param url_string: URL (string) of this result.
        @param type_name: Type of the result as a string.
        @param method: Method with which the result was obtained.
        @return: Result object with valid type but wihout meta information.
        """

        rtype_id = static_result_type_id(type_name)
        if rtype_id == 1:
            logger.warning("Added unknown result type %s " % type_name)
        if DEBUGANALYSER:
            logger.debug("Adding result:\nURL: %s\nType: %s\nMethod:%s" %\
                    (url_string, type_name, method))
        result = Result(url_string, method=method)
        result.result_type_id = rtype_id
        result.date = date

        return result

    def commit(self):
        """
        Commits the results of the analyser to the database and adds
        metadata which is only available after a full scan.
        """
        for result in self.visit.results:
            if result.date == None:
                result.date = self.visit.date
            Session.add(result)
            if result.result_file:
                Session.add(result.result_file)
        Session.add(self.visit)
        Session.commit()

    def is_interesting(self):
        """
        Check if the results are interesting.

        @return: True if something interesting has been found.
        """
        for result in self.visit.results:
            # TODO Just access a proper index (id->severity) in memory.
            if static_result_type_severity(result.result_type_id) >= self.tolerance:
                return True
        logging.debug("Have a visit which is not interesting.")
        return False

    def archive_or_delete(self):
        """
        Removes file with uninteresting results, dump etc.
        """
        if not self.is_interesting():
            os.remove(self.visit.filepath)
            self.visit.filepath = None

    def generic_analyse(self):
        """
        Runs generic steps of the analysis.
        @return: True if a result has been found, False otherwise.
        """
        if self.visit.results:
            self.visit.interesting = self.is_interesting()
            # Do not add results to db if visit is not interesting.
            if not self.visit.interesting:
                for result in self.visit.results:
                    if result in Session:
                        Session.expunge(result)
                    del result
                self.visit.results = []
            self.archive_or_delete()
            self.commit()
            logging.debug("Added new visit into database with %d results." %\
                           len(self.visit.results))
            return True

        logging.warning("Sucessfully scanned %s but did not get any results."%\
                        self.result_data)
        return False

# TODO: This class is probably obsolete, remove it.
class TarballAnalyser(ResultAnalyser):
    """
    Class to analyse tarballs of Results provided
    by the mapwocanalyse.pl script.
    """

    def __init__(self, result_data, node):
        """
        Resultanalyser class to enter raw result_data into the Database as
        results.

        @param result_data: Qualifier (path to) of the raw data which should be parsed.
        @param node: Name of node on which the scan was running.
        """
        ResultAnalyser.__init__(self, result_data, node)

    def analyse_dry(self):
        """For use only in debugging and development environments."""
        self.__parse_tarball(self.result_data)
    
    def analyse(self):
        """
        Starts the analysis of a result.
        @return: True if at least one result could be entered into
                the database.
        """
        try:
            self.__parse_tarball(self.result_data)
            return self.generic_analyse()
        except:
            traceback.print_exc()
            logging.error("Invalid tarball %s ignoring result." % \
                    self.result_data)
            Session.rollback()
            return False

    def __parse_tarball(self, resultb):
        """
        Imports a result archive into the db.

        @param resultb: Path to tarball with results.
        """

        if DEBUGANALYSER:
            logger.debug("Parsing result- tarball %s ." % resultb)
        tgz = tarfile.open(resultb)

        # Call function [1] with contents of file [0] .
        file_func_map = [('result/Log/header.log', self.__scan_headers),
                         ('result/Log/302.log', self.__scan_302),
                         ('result/Log/Test.log', self.__scan_testlog),
                         ('metainfo', self.scan_metainformation),
                         ('result/Log/404.log', self.__scan_404),
                         ('result/Log/Analyse.log', self.__scan_analyse)]

        if DEBUGANALYSER:
            logger.debug("Files in tarball: %s" % tgz.getnames())

        tgz_names = tgz.getnames()
        # Delegate interesting files to subfunctions.
        for (anl_file, anl_func) in file_func_map:
            if anl_file in tgz_names:
                anl_func(tgz.extractfile(anl_file))

        if DEBUGANALYSER:
            logger.debug("URLs: %s" % ", ".join([url.url for url in self.visit.urls]))

        for fname in [finame for finame in tgz_names if finame.startswith('result/Binary')]:
            if fname.startswith("result/Binary/GOOGLE"):
                # TODO Could also be evil
                continue
            self.__scan_binary(tgz.extractfile(fname))
            continue

    def __scan_testlog(self, log):
        """ Scan the TestLog and extract the hcsid from the dump file name. """
        hc = log.readlines()[0].rsplit("/", 1)[1].strip()
        # Remove the 6 tmp chars, _ and .dump and snapshot
        self.visit.hcsid = hc[:len(hc)-21]
        logger.debug("Found Test.log extracted hcsid: %s " % self.visit.hcsid)

    def __scan_binary(self, fname):
        pass

    def __scan_analyse(self, log):
        """ Scan the Analyse.log file and modify visit accordingly. """

        # Remove duplicate lines.
        lines = set(log.readlines())
        method = "mapwocanalyse.pl/Analyse.log/0.0.2"

        image_types = ["[GIF]", "[PNG]", "[JPEG]"]
        nr_images = 0

        # TODO understand the structure of this file
        # and why there are duplicates of the same line
        # For now just add one result
        for line in lines:
            if DEBUGANALYSER:
                logging.debug("__scan_analyse: line %s " % line)
            recs = line.strip().split()
            if not len(recs):
                continue
            analyse_type = recs[0]
            if analyse_type == "[GOOGLE]":
                host = recs[3]
                referer = recs[4]

                # When browsing manually we do not have an URL (that is in db)
                if self.visit.urls and len(self.visit.urls) and self.visit.urls[0]:
                    url_str = self.visit.urls[0].url
                else:
                    url_str = "manual browsing"
                result = self.construct_result(url_str, "GoogleAnalytics",
                                               method)

                result.text = "%s\n%s" % ( host, referer )
                self.visit.results.append(result)
            elif analyse_type == "[EXE]":
                host = recs[3]
                referer = recs[4]
                path = recs[-1]
                #Example of a line:
                #[EXE]
                #../10.0.0.4_49159_80
                #110
                #files.kolab.org
                #widgettest.exe
                #files.kolab.org/incoming/aheinecke/
                #(REFERER) ...

                url_str = recs[5] + recs[4]

                result = self.construct_result(url_str,
                                               "WindowsExecutable",
                                               method)

                # TODO Values are not what they are named
                result.text = ("Windows PE32 found at %s."
                        " (from %s\n, ~Referer: %s)" % ( url_str, host, referer ))
                if DEBUGANALYSER:
                    logger.debug(result.text)
                resultfile = ResultFile()
                resultfile.filepath = recs[4]
                # TODO Calculate sha1sum of file content (file recs[2]?).
                resultfile.sha1sum = 'differs'
                #util.sha1sum(recs[-2])
                resultfile.ssdeep = "ssdeep sum"
                result.result_file = resultfile
                self.visit.results.append(result)
            elif analyse_type in image_types:
                nr_images = nr_images + 1
            else:
                host = recs[3]
                referer = recs[4]

                # When browsing manually we do not have an URL (that is in db)
                if self.visit.urls and len(self.visit.urls) and self.visit.urls[0]:
                    url_str = self.visit.urls[0].url
                else:
                    url_str = "manual browsing"
                result = self.construct_result(url_str, "unknown",
                                               method)

                result.text = "Analyse-Type: %s\n" % analyse_type
                self.visit.results.append(result)
        if nr_images > 0:
            # When browsing manually we do not have an URL (that is in db)
            if self.visit.urls and len(self.visit.urls) and self.visit.urls[0]:
                url_str = self.visit.urls[0].url
            else:
                url_str = "manual browsing"
            result = self.construct_result(url_str, "GraphicsFound",
                                           method)
            result.text = "%d images found." % ( nr_images )
            self.visit.results.append(result)


    def __scan_headers(self, log):
        """ Scan the Headers.log file and add Results accordingly. """

        # TODO Making hard assumptions (http)
        lines = log.readlines()
        if len(lines) <= 1:
            # Empty header log, TODO does this mean url can not be resolved?
            logger.debug("Empty header log. Adding MissingTraffic result.")
            result = self.construct_result('n/a', 'MissingTraffic', 'mapwocanalyse.pl/header.log/0.0.2')
            result.text = "Log/header.log was empty."
            self.visit.results.append(result)
            return
        else:
            # Fields in line: src-ip,time,dst(host),agent(crippled) .
            # If this is not the case something is broken.
            try:
                # TODO take this from redirection log
                #url_string = "http://%s" % lines[2].split()[2]
                #url = Session.query(URL).filter(URL.url == url_string).first() or \
                #      Session.query(URL).filter(URL.url == url_string.rstrip('/')).first()
                #self.visit.urls.append(url)
                #Session.rollback()
                for line in lines[2:]:
                    if DEBUGANALYSER:
                        logger.debug("__scan_headers: line: %s " % line)
                    recs = line.strip().split()
                    url_string = "http://%s" % recs[2]
                    result = self.construct_result(url_string, "RegularTraffic",
                            "mapwocanalyse.pl/header.log/0.0.2")
                    result.text = '- found in Log/header.log -'
                    self.visit.results.append(result)
                    if DEBUGANALYSER:
                        logger.debug("Added RegularTraffic result")
            except:
               result = self.construct_result("n/a", "MissingTraffic",
                                              "mapwocanalyse.pl/header.log/0.0.2")
               self.visit.results.append(result)
               if DEBUGANALYSER:
                   logger.debug("Something went wrong. Added MissingTraffic result.")
                   traceback.print_exc()


    def __scan_404(self, log):
        """ Parses a 404.log and creates the relevant results. """

        for line in log.readlines():
            url_string = line.split('\t')[-1].strip()
            result = self.construct_result(url_string, "404",
                            method='mapwocanalyse.pl/404/0.0.2')
            result.text = line.strip()
            self.visit.results.append(result)
            logger.debug("Added 404 result")

    def __scan_302(self, log):
        """Parses a 302.log and registers the URLs from redirector."""

        # TODO 302s can be interesting results, too.
        for line in log.readlines():
            records = line.split()
            # TODO Remove hard-coded redirector address.
            logger.debug("302: %s" % line)
            logger.debug("302 - r3 : %s" % records[-2])

            if records[3] == "10.0.0.1":
                # Found our redirector.
                # This URL shall be in database.
                # TODO use sqlalchemys _or  .
                url = Session.query(URL).filter(URL.url == records[-2]).first() or \
                      Session.query(URL).filter(URL.url == records[-2].rstrip('/')).first()
                if url:
                    self.visit.urls.append(url)
                if DEBUGANALYSER:
                    logger.debug("Added an URL from 302.log.")

