# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process to poll results from nodes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import glob
import logging
import multiprocessing
from Queue import Empty
import tempfile
from time import sleep
from os import listdir, path, walk, rmdir
from os.path import isdir
import subprocess
import tarfile
import datetime

from sqlalchemy import *

from mapwoc.lib.result import Result, ResultType, Visit
from mapwoc.lib.url import URL
import mapwoc.lib.model.result
import mapwoc.lib.model.url
from mapwoc.lib.tickprocess import TickProcess
from mapwoc.masterctrl.resultfileanalyser import ResultFileAnalyser

from mapwocgui.model.meta import Session

logger = logging.getLogger('master')

class ResultPollProcess(TickProcess):
    """
    Process that polls for results (rsnyc) of nodes.
    In case that any results where fetched, these are registered
    in the database.
    """

    def __init__(self, working_dir, nodes, result_queue, credentials_path, db,
                 tolerance):
        """
        Constructor, set up fields to sane values.

        @param working_dir: Directory in which to create temporary dir. for syncing.
        @param nodes: NodeConnection list of nodes with which to sync.
        @param result_queue: Queue in which to write when results were found.
        @param credentials_path: Path to certificates for ssh/rsync
        @param db: Path to database in which to store the results.
        @param tolerance: Required highest severity for a result to store.
        """

        TickProcess.__init__(self, result_queue, 10)
        self.__working_dir = working_dir
        self.__nodes = nodes
        self.credentials_path = credentials_path
        self.db_file = db
        self.tolerance = tolerance

    def __bind_to_db(self):
        """Bind the Session to database."""

        # TODO Find out whether sqlalchemy can protect from concurrency issues,
        #      and how.
        engine = create_engine('sqlite:///%s' % self.db_file) 
        c = engine.connect() 
        Session.configure(bind=engine)

    def tick(self):
        """
        Called upon call to start().
        Starts a Rsync process peridically.
        """

        self.__bind_to_db()

        rsnc_subprocess = None

        # Call that rsync subprocess, sync from there into:

        for nodekey, node in self.__nodes.items():
            tempd = path.abspath(tempfile.mkdtemp(dir=self.__working_dir,
                                       prefix='resultsync_%s_' % nodekey))
            if not node.data_path:
                logger.error(
                        "Node %s does not have a resultdir set." +\
                        " Is it running?", nodekey)
                continue
            cmd = ['rsync', '-az',
                    '-e',
                    ('"/usr/bin/ssh -oKbdInteractiveAuthentication=no '
                      '-oPasswordAuthentication=no '
                      '-oNumberOfPasswordPrompts=0 -i %s"'
                    ) % path.abspath(self.credentials_path),
                    '--remove-source-files',
                    '%s:%s' % (node.rsync_address, node.data_path),
                    tempd ]
            p = subprocess.Popen(" ".join(cmd),
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            shell=True)
            p.wait()

            if p.returncode != None:
                if p.returncode == 0:
                    logger.info("Result Sync with node %s successful."
                                % nodekey)
                    # Now, check what came and import it to db.
                else:
                   stdout_value, stderr_value = p.communicate()
                   logger.warning("Could not sync results with node %s."
                                  % nodekey)
                   logger.debug("stdout: %s" % stdout_value)
                   logger.debug("stderr: %s" % stderr_value)
                   logger.debug("exit code: %d" % p.returncode)
                   logger.debug("command was: %s" % cmd)
            else:
                logger.warning("Rsync without returncode.")

            # Walk over rsynced once, put in db/remove.
            # TODO: Signal the master if something serious is happening.
            if self.__resultdir_to_db(tempd, nodekey):
                self.put('marker')
            #(self.master.update_results)


    def __resultdir_to_db(self, resultdir, node=None):
        """
        Checks a result directory for result tarballs
        and calls the according analayser.
        """

        for dirpath, dirnames, filenames in walk(resultdir):
            if not filenames and not dirnames:
                # clean up empty directories also removes an empty
                # resultdir
                rmdir(dirpath)
                continue
            for filename in filenames:
                tarpath = path.join(dirpath, filename)
                if not filename.endswith(".tar.gz"):
                    logger.warning("Analysis: Unkown file: %s ." \
                                   % tarpath)
                else:
                    analyser = ResultFileAnalyser(tarpath, node, self.tolerance)
                    if not analyser.analyse():
                        logger.warning("Could not extract result from %s ."\
                                        % tarpath)
