# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Analyser-Process related classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import multiprocessing
import logging
from os import remove
from os.path import basename, abspath
from shutil import copy, move
import subprocess
import tarfile
import tempfile
from time import sleep

from mapwoc.lib.result import Result

logger = logging.getLogger("node")

class AnalysisProcess(multiprocessing.Process):
    """
    Analysis Process sits and wait for incoming filepaths in job_queue.
    Once, a filepath could be fetched from this queue, the analysis script
    is thrown at it and the results path is fed ("back") into the result_queue.
    """


    def __init__(self, job_queue, result_queue, working_dir, result_dir,
                 analyser = '/usr/bin/mapwocanalyse.pl'):
        """
        Constructor.
        @param job_queue: Queue for (dump-)filenames to be analysed.
        @param result_queue: Queue for results to be put when analysis finished.
        @param working_dir: Working directory (where temp. files will be placed.)
        @param result_dir: Directory where result tarballs will be placed.
        @param analyser: Path to the analyser which should be used.
        """

        multiprocessing.Process.__init__(self)
        self.result_queue = result_queue
        self.job_queue = job_queue
        self.working_dir = abspath(working_dir)
        self.result_dir = abspath(result_dir)
        self.analyser = analyser

    def run(self):
        """
        Invoked on start() - call.
        Polls Queue for (dump-)filenames and analyses these if possible.
        Paths to results are pushed in the result queue.
        """

        while True:

            job = self.job_queue.get()

            # Compatibility code for old tests
            if type(job) != type(tuple()):
                tcpdump = job
                proxydump = None
            else:
                tcpdump = job[0]
                proxydump = job[1]

            logger.debug("Starting analysis of: %s" % tcpdump)

            tempdir = abspath(
                    tempfile.mkdtemp(dir=self.working_dir, prefix='anaylyser_'))
            # TODO When copying fails, node should go into error state.
            copy(tcpdump, tempdir)
            if proxydump:
                copy(proxydump, tempdir)

            # Obtain a temporary working dir.
            cmd = [self.analyser, "-p", abspath(tcpdump), "-b", tempdir]

            if proxydump:
                cmd += ["-j", abspath(proxydump)]

            process = subprocess.Popen(cmd, close_fds=True,
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            # Wait for process to finish.
            process.wait()
            if True or process.returncode:
                stdout_value, stderr_value = process.communicate()
                logger.debug("stdout: %s" % stdout_value)
                logger.debug("stderr: %s" % stderr_value)
                logger.debug("exit code: %d" % process.returncode)

            # Now bundle the results.
            _fh,fname = tempfile.mkstemp(dir=self.working_dir, suffix='.tar')
            tgz = tarfile.open(name=fname, mode='w')
            tgz.add(tempdir, arcname='result')
            tgz.close()

            remove(tcpdump)
            if proxydump:
                remove(proxydump)
            self.result_queue.put(fname)
