# -*- coding: utf-8 -*-

# mapWOC
#
# Description: KVM-Process related classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import multiprocessing
from os import path, remove, chmod, makedirs
from shutil import copy
import socket
import stat
import subprocess
from tempfile import mkstemp
from time import sleep
from Queue import Empty

from mapwoc.lib.hc import HC
from mapwoc.nodectrl.hcinteractionscript import HCInteractionScript
from mapwoc.nodectrl.hcrunprocess import HCRunProcess

logger = logging.getLogger("node")

class KVMHCRunProcess(HCRunProcess):
    """
    Process that starts a VM in a subprocess and talks to it via stdin.
    """

    SLEEP_STEP = 1

    def __init__(self, hc, script, timeout, command_queue, result_queue,
                 work_dir, initial_offset=0, proxy_port=1234):
        """
        Constructor.
        @param hc: An HCControl object.
        @param script: A script to be executed (list of pairs (second,action)).
        @param timeout: When the subprocesses will be terminated, in seconds.
        @param command_queue: Regularly polled by process, if non-empty terminate.
        @param result_queue: When process is finished, but path to the dump file.
        @param work_dir: Where the dump file will be placed.
        @param initial_offset: Time in seconds to sleep before starting the
                               process (when start() or run() is called).
        @param proxy_port: The port on which the https proxy shoul be started.
        """

        HCRunProcess.__init__(self, timeout=timeout, hc=hc)
        
        self.script = HCInteractionScript(script)
        self.cp_img_path = ''
        self.vnc_port = None
        self.proxy_port = proxy_port
        self.dump_file = ''
        self.kvm_subprocess = None
        self.proxy_subprocess = None
        self.work_dir = work_dir
        self.__prepare_kvm_files()
        self.initial_offset = initial_offset
        self.waiting_for_start = True
        self.result_queue = result_queue
        self.command_queue = command_queue

    def __del(self):
        """Destructor."""

        # TODO ensure files (from __prepare_kvm_files) are removed.
        pass

    def __prepare_kvm_files(self):
        """Prepare files and paths (e.g. copy image file)."""

        if not path.isdir(self.work_dir):
            makedirs(self.work_dir)

        _handle, self.dump_file = mkstemp(prefix=path.basename('%s_' %
                                          self.hc.hcsid),
                                          suffix='.dump',
                                          dir=self.work_dir)

        self.proxy_dump = self.dump_file.replace(".dump", ".proxy")
        self.cp_img_path = self.hc.image_file

    def __kvm_cmd(self):
        """
        Create KVM Command to start the given HC.
        @return: List to be used with Popen to start kvm process.
        """

        # Networking parameters.
        usermode_net_params = ''
        # Proxy-forward.
        if self.hc.proxy_ip:
            usermode_net_params += ',guestfwd=tcp:%s:80-cmd:/bin/nc.traditional localhost %s' % (self.hc.proxy_ip,self.proxy_port)
        # Network-related.
        if self.hc.own_ip:
            usermode_net_params += ',net=%s' % self.hc.own_ip
        if self.hc.dns_ip:
            if not usermode_net_params.endswith(','):
                usermode_net_params += ','
            usermode_net_params += 'dns=%s' % self.hc.dns_ip
        if self.hc.gateway_ip:
            if not usermode_net_params.endswith(','):
                usermode_net_params += ','
            usermode_net_params += 'host=%s' % self.hc.gateway_ip
        nic_model_params = ''
        if self.hc.hctype == 'kvm-mac':
            nic_model_params = ',model=e1000'

        # Do live migration if not booting.
        migration_cmds = []
        if not self.hc.boots:
            migration_cmds.append('-incoming')
            snapshot_path = self.hc.image_snapshot
            # Is the snapshot (state) compressed?
            if snapshot_path.endswith('gz'):
                migration_cmds.append('exec:gzip -dc  %s' %  snapshot_path)
            else:
                migration_cmds.append('exec:cat %s' % snapshot_path)

        # TODO We could drop privileges (to privs. of other user) with -runas .
        # Maybe kvm should be started as --daemonize.

        cmd = [ "/usr/bin/kvm",
                "-hda", self.cp_img_path,
                "-net",
                # All -net user options need to be made in the following line
                "user%s" % usermode_net_params, # Use user mode network
                "-net", "nic%s" % nic_model_params, # Create Network Interface Card in VLAN 0
                "-net", "dump,file=%s" % self.dump_file # Dump
              ] +  migration_cmds + [ # VM State to load.
                '-snapshot', # Do not alter files.
                '-m', str(self.hc.image_memory),
                '-rtc', "base=localtime",
              ]
        if self.vnc_port != None:
             cmd += [ "-vnc",  ":%d" % self.vnc_port ]
        else:
             # TODO '-nographic' should save ressources but proved to be
             #      difficult to insert (depends on position in kvm cmd).
             #cmd.append('-nographic')
             cmd += [ "-vnc",  "none" ]
        if self.hc.hctype == 'kvm-mac':
            cmd += [
                    '-M', 'mac',
                    '-osk', 'ourhardworkbythesewordsguardedpleasedontsteal(c)AppleComputerInc',
                    '-no-kvm-irqchip',
                    '-usbdevice', 'keyboard',
                    '-usbdevice', 'mouse'
                   ]
        if self.hc.kvmkernel:
            cmd += [
                    '-kernel', self.hc.kvmkernel
                   ]
        cmd += [ "-monitor", "stdio"]

        return cmd

    def run(self):
        """
        Invoked on start() - call.
        Starts a KVM subprocess, a man in the middle proxy for ssl traffic
        and a parallel interacts with the kvm process by sending keyboard-press
        commands from a script to the kvm monitor.
        After timeout, abort the processes.
        """

        cmd = self.__kvm_cmd()

        if self.initial_offset > 0:
            logger.debug("HC with offset started, sleeping %f" % float(self.initial_offset))
            sleep(self.initial_offset)
        self.waiting_for_start = False

        # Start mitmproxy subprocess.
        # This will raise an error if something goes wrong.
        if self.hc.proxy_ip:
            # TODO Proxy-functionality is also needed for native clients.
            logger.debug("Starting mitmdump on port: %s" % self.proxy_port)
            # TODO: Register the request-script.
            mitmproxy_cmd = ["mitmdump", "-v", "-p", str(self.proxy_port),
                             "--reqscript",
                             "/home/felix/mapwoc/proxy/mapwoc_redirect_script",
                             "-w", self.proxy_dump]
            self.proxy_subprocess = subprocess.Popen(mitmproxy_cmd,
                                                     close_fds=True)
            sleep(1) # Give the proxy time to open the port

        # Start kvm subprocess.
        logger.debug("Starting HC with: %s" % cmd)
        self.kvm_subprocess = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE,
                                               close_fds=True)

        # Init time.
        self.start_time()

        while not self.time_over():
            self.kvm_subprocess.poll()
            if self.kvm_subprocess.returncode is not None:
                # TODO escalate this
                stdout_value, stderr_value = self.kvm_subprocess.communicate()
                logger.warning("KVM Process seemed to have died all by itself.")
                logger.warning("stderr: %s" % stderr_value)
                logger.warning("stdout: %s" % stdout_value)
                logger.warning("exit code: %d" % self.kvm_subprocess.returncode)
                break

            try:
                cmd = self.command_queue.get_nowait()
                # If command is given, this is the command to stop.
                break
            except Empty:
                pass

            while self.script.needs_action(self.elapsed_time()):
                action = self.script.next_action()
                self.kvm_subprocess.stdin.write('sendkey %s\n\r' % action)
                self.kvm_subprocess.stdin.flush()
                self.script.action_done()
                logger.debug("Sent keypress (%s) to VM." % action)

            sleep(KVMHCRunProcess.SLEEP_STEP)

        # Terminate the Processes
        if self.hc.proxy_ip:
            self.__terminate(self.proxy_subprocess)
        if self.__terminate(self.kvm_subprocess) == None:
            # TODO Ensure that the process is joined.
            pass
        self.cleanup()
        if self.hc.proxy_ip:
            self.result_queue.put((self.dump_file, self.proxy_dump))
        else:
            self.result_queue.put(self.dump_file)


    def __terminate(self, process):
        """
        Terminates the kvm subprocess and waits for it to
        shut down.  Blocks for up to 10 seconds before
        sending sigkill.

        @return: The subprocesses return code.
        """

        state = process.poll()
        # TODO Send quit command to process.
        #process.stdin.write('quit\n\r')
        if state == None:
            process.terminate()
        waited = 0
        while state == None and waited < 10:
            state = process.poll()
            if state:
                break
            sleep(1)
            waited += 1
        if state == None:
            logger.warning("subprocess not terminated after 10 seconds.")
            process.kill()
        return state

    def stop_processes(self):
        """Big red button."""
        # TODO examine issue why this doesnt work
        # (probably related to process/forking)

        self.command_queue.put('stop now')
        if self.kvm_subprocess is not None:
            self.kvm_subprocess.terminate()
        if self.kvm_subprocess is not None:
            self.kvm_subprocess.kill()
        if self.proxy_subprocess is not None:
            self.proxy_subprocess.terminate()
        if self.proxy_subprocess is not None:
            self.proxy_subprocess.kill()

    def stop(self):
        """Stop processes directly."""

        self.stop_processes()

    def get_result(self):
        """
        Get result(s) from queue.

        @return: Pair with dumpfiles if proxy was configured, single dump
                 file path otherwise.
        """

        return self.result_queue.get_nowait()

    def cleanup(self):
        """Any cleanups to do (none, currently)."""

        #remove(self.dump_file)
        #remove(self.proxy_dump)

        logger.debug('Cleaning up kvm process, leaving dump file.')

