import logging

from pylons import request, response, session, tmpl_context as c, url
from pylons.controllers.util import abort, redirect
from pylons.i18n.translation import _, ungettext

import formencode

from pylons.decorators import validate
from pylons.decorators.rest import restrict
from sqlalchemy import func
import webhelpers.paginate as paginate

from mapwocgui.lib.base import BaseController, render
import mapwocgui.lib.helpers as h 

from mapwocgui.lib.auth import checkAccess
from mapwocgui.model import Session
from mapwoc.lib.url import URLList, URL, create_URL_List, separate_valid_urls
from mapwoc.lib.model.url import url_table, urllist_url, urllist_table, insert_list

log = logging.getLogger(__name__)

class NewURLListForm(formencode.Schema):
    """URLList Validation, dummy."""

    # Extra field is submit button, right now
    allow_extra_fields = True
    filter_extra_fields = True
    URLListName = formencode.validators.String(not_empty=True)
    URLListComment = formencode.validators.String()
    URLsText = formencode.validators.String(not_empty=True)

class UploadURLListForm(formencode.Schema):
    """URLList upload validation."""

    # Extra field is submit button, right now
    allow_extra_fields = True
    filter_extra_fields = True
    URLListUploadName = formencode.validators.String(not_empty=True)
    URLListUploadComment= formencode.validators.String()
    URLListFile = formencode.validators.String(not_empty=True)

class UrllistController(BaseController):

    def index(self):

        # Do not coun in ORM using len(urllist.urls), but speed things up
        # a little using a custom query.
        q = Session.query(URLList, func.count())
        q = q.outerjoin((urllist_url, urllist_url.c.urllist_id == URLList.id))
        q = q.group_by(URLList.id)
        q = q.order_by(URLList.name)

        c.urllist_paginator = paginate.Page(q,
                                   page=int(request.params.get('page',1)),
                                   items_per_page=15,)
        return render('/urllist/index_urllist.mako')

    @checkAccess("privileged")
    def new(self):
        return render('/urllist/index_urllist.mako')

    def __handle_invalids(self, invalids):
        """ Sets a flash alert for invalid URLs """

        if not invalids:
            return
        if len(invalids) < 20:
            h.flash(_(u"URL list contained invalid URLs:"))
            for invalid in invalids:
                h.flash(invalid)
        else:
            h.flash(_(u"More then 20 URL's in the list were invalid."))
            h.flash(_(u"Please check the created list."))

    @checkAccess("privileged")
    @restrict('POST')
    @validate(schema=UploadURLListForm(), form='new')
    def upload(self):
        """ Creates an URLList from a file. """

        filecontent = request.POST['URLListFile'].value

        valids,invalids = separate_valid_urls(filecontent)
        self.__handle_invalids(invalids)

        # Create from request.
        urllist_name = self.form_result['URLListUploadName']
        urllist_comment = self.form_result['URLListUploadComment'] or ''

        # TODO detect and respect fail cases.
        uid = insert_list(Session, set(valids), urllist_name, urllist_comment)

        Session.flush()
        Session.commit()

        h.good_flash(_("URL-List created."))
        response.status_int = 302
        response.headers['location'] = h.url_for(controller='urllist',
                                                 action='index')
        return "Moved temporarily"


    @checkAccess("privileged")
    @restrict('POST')
    @validate(schema=NewURLListForm(),form='new')
    def create(self):
        """Creates an URLList from a request."""

        # Create from request
        urllist = URLList(self.form_result['URLListName'])
        urllist.comment = self.form_result['URLListComment'] or ''

        valids,invalids = separate_valid_urls(self.form_result['URLsText'])
        self.__handle_invalids(invalids)

        # Avoid duplicate URLs in database.
        # TODO Note optimization potential here.
        for url_str in valids:
            url = Session.query(URL).filter_by(url=url_str).first()
            if not url:
                # URL not yet in database.
                url = URL(url_str)
                Session.add(url)
                Session.flush()
            urllist.urls.append(url)

        # Add to store
        Session.add(urllist)
        Session.commit()

        # Redirect
        # TODO use redirect_to from pylons.controller.utils ?
        response.status_int = 302
        response.headers['location'] = h.url_for(controller='urllist', action='index', id=urllist.id)
        return "Moved temporarily" 

    def view(self, id):
        query = Session.query(URLList)
        c.urllist = query.get(int(id))

        # Replace c.urllist.urls by a query for perfomance reasons.
        q = Session.query(URL).filter(URL.urllist.any(id=id))
        c.url_paginator = paginate.Page(q,
                        page=int(request.params.get('page',1)),
                        items_per_page=100,)
        return render('/urllist/view_urllist.mako')

    #TODO check whether the url list is used in any scans.
    @checkAccess("privileged")
    def delete(self, id):
        if id is None:
            abort(404)
        urllist_query = Session.query(URLList)
        urllist = urllist_query.filter_by(id=id).first()
        if urllist is None:
            abort(404)
        urllist.delete()
        Session.commit()
        response.status_int = 302
        response.headers['location'] = h.url_for(controller='urllist',action='index',id=urllist.id)
        return "Moved temporarily"        

